/*
 *  message.h - Types and constants used for message-passing.
 *
 *  (C) 1994 Mikael Nordqvist (d91mn@efd.lth.se, mech@df.lth.se)
 */

#ifndef _MESSAGE_H
#define _MESSAGE_H

#define MESSAGE_JUMPPATTERN   1
#define MESSAGE_RESTART       2
#define MESSAGE_CHANGESPEED   3
#define MESSAGE_STOP          4

#define MESSAGE_PRINTLINE    10
#define MESSAGE_PRINTSPEED   11

#define MESSAGE_DONE         98
#define MESSAGE_EXIT         99

#define MESSAGE_NOP         120

struct mod_message {
    char type;
    
    union {
	unsigned char songpos;
	unsigned char speed;
    } a1;
    
    union {
	unsigned char line;
	unsigned char tempo;
    } a2;
};

#endif /* _MESSAGE_H */
