/*
 *  help.h - Help and usage texts.
 *
 *  (C) 1994 Mikael Nordqvist (d91mn@efd.lth.se, mech@df.lth.se)
 */

#define USAGE_TEXT1 "\
mod version 0.6 Copyright (C) Mikael Nordqvist, 1994.\n\n\
Usage: %s [global options] file [options] file [options] ...\n"

#define USAGE_TEXT USAGE_TEXT1 "\
Options: [aAbBcCfhklLmMnNoOpPqQrstvV056]\n\n\
Use %s -h for help.\n"

#define HELP_TEXT USAGE_TEXT1 "\
\n\
Options always global:\n\
\n\
  -a        Enable auto-next             -l <arg>  Number of scroll-lines\n\
  -A        Disable auto-next            -n        Disable scrolling\n\
  -b        Break loops                  -q        Quiet mode\n\
  -B        Allow loops                  -Q        Totally quiet mode\n\
  -C        Cyclic play of modules       -r        Play modules randomly\n\
  -h        Help                         -v        Verbose mode\n\
  -k        Kill mod in background       -z        Put mod in background\n\
\n\
Module-specific options:\n\
\n\
  -c <arg>  Click removal                -p <arg>  Start-position\n\
  -f <arg>  Module format                -P        PAL-amiga samplepitch\n\
  -L        Loop module                  -s <arg>  Set speed\n\
  -m <arg>  Maximum moduletime           -t <arg>  Set tempo\n\
  -M        Mono output                  -x <arg>  Detail of voice-information\n\
  -N        NTSC-amiga samplepitch       -0        Ignore speedchanges to 0\n\
  -o        Octaves 1-3                  -5        PAL-amiga (50 Hz) tempo\n\
  -O        Octaves 0-7                  -6        NTSC-amiga (60 Hz) tempo\n\
"
